# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from . import menu
from . import view
from . import icon

__all__ = ['menu', 'view', 'icon']
